/*
 * Chess game implementation.
 * By Carlos Varela.  v0.1  February, 1997
 *
 * Board class.  Contains the chess board.
 */

package chess;

import java.util.*;

public class Board implements submit.Board, java.io.Serializable  {

  Game game;
  static final char[] column = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'};
  Piece[][] boardMatrix;
  Vector whitePiecesOnBoard, blackPiecesOnBoard;
  Vector whitePiecesOut, blackPiecesOut;
  Piece bKing, wKing;
  boolean castling = true;  // Whether to generate castling as a move or not.
  Vector moves;             // Vector of moves to get to this board.

  public Board() {
    this(new Game());
  }
  
  public Board(String position) {
    this(new Game(), position);
  }
  /*
   * Creates an initial chess board.
   */
  Board(Game g){
    game = g;
    whitePiecesOnBoard = new Vector(16);
    blackPiecesOnBoard = new Vector(16);
    whitePiecesOut = new Vector(16);
    blackPiecesOut = new Vector(16);
    boardMatrix = new Piece[8][8];           //  Set the pieces in the board.

    for (int i=0; i<8; i++){
      setPieceAt(column[i], 2, new Pawn(Game.WHITE));       // Pawns
      setPieceAt(column[i], 7, new Pawn(Game.BLACK));
    }

    setPieceAt('a', 1, new Rook(Game.WHITE));
    setPieceAt('h', 1, new Rook(Game.WHITE));              // Rooks
    setPieceAt('a', 8, new Rook(Game.BLACK));
    setPieceAt('h', 8, new Rook(Game.BLACK));

    setPieceAt('b', 1, new Knight(Game.WHITE));
    setPieceAt('g', 1, new Knight(Game.WHITE));             // Knights
    setPieceAt('b', 8, new Knight(Game.BLACK));
    setPieceAt('g', 8, new Knight(Game.BLACK));

    setPieceAt('c', 1, new Bishop(Game.WHITE));
    setPieceAt('f', 1, new Bishop(Game.WHITE));             // Bishops
    setPieceAt('c', 8, new Bishop(Game.BLACK));
    setPieceAt('f', 8, new Bishop(Game.BLACK));

    setPieceAt('d', 1, new Queen(Game.WHITE));              // Queens 
    setPieceAt('d', 8, new Queen(Game.BLACK));

    setPieceAt('e', 1, wKing = new King(Game.WHITE));       // Kings
    setPieceAt('e', 8, bKing = new King(Game.BLACK));

    for (int i=0; i < 8; i++){
      whitePiecesOnBoard.addElement(getPieceAt(column[i],1));
      whitePiecesOnBoard.addElement(getPieceAt(column[i],2));
    }
    for (int i=0; i < 8; i++){
      blackPiecesOnBoard.addElement(getPieceAt(column[i],7));
      blackPiecesOnBoard.addElement(getPieceAt(column[i],8));
    }
    moves = new Vector(80);
  }

  /*
   * Creates an initial chess board with this game and
   * a given position in the format:
   *
   * r3nrk1/p3qppp/bp2pn2/2pp4/2PP2N1/P1PBPP2/1BQ3PP/1R3RK1
   *
   * for a board like:
   * 
   * [--------------
   * r . . . n r k .
   * p . . . q p p p
   * b p . . p n . .
   * . . p p . . . .
   * . . P P . . N .
   * P . P B P P . .
   * . B Q . . . P P
   * . R . . . R K .
   * white to play
   * --------------]
   *
   */
  Board(Game g, String position){
    game = g;
    whitePiecesOnBoard = new Vector(16);
    blackPiecesOnBoard = new Vector(16);
    whitePiecesOut = new Vector(16);
    blackPiecesOut = new Vector(16);
    boardMatrix = new Piece[8][8];           //  Set the pieces in the board.

    char column = 'a';
    int row = 8;
    Piece p = null;

    for (int i=0; i<position.length();i++){
      char c = position.charAt(i);
      if (c == '/'){
	row--;
	column = 'a';
      } else if ((c >= '1') && (c <='8')){
	column = (char) (column + (c - '0'));
      } else {
	switch (c){
	case 'k': 
	  p = bKing = new King(Game.BLACK); break;
	case 'q':
	  p = new Queen(Game.BLACK); break;
	case 'r':
	  p = new Rook(Game.BLACK); break;
	case 'b':
	  p = new Bishop(Game.BLACK); break;
	case 'n':
	  p = new Knight(Game.BLACK); break;
	case 'p':
	  p = new Pawn(Game.BLACK); break;
	case 'K': 
	  p = wKing = new King(Game.WHITE); break;
	case 'Q':
	  p = new Queen(Game.WHITE); break;
	case 'R':
	  p = new Rook(Game.WHITE); break;
	case 'B':
	  p = new Bishop(Game.WHITE); break;
	case 'N':
	  p = new Knight(Game.WHITE); break;
	case 'P':
	  p = new Pawn(Game.WHITE); break;
	}
	setPieceAt(column,row,p);
	if (p.color())
	  whitePiecesOnBoard.addElement(p);
	else
	  blackPiecesOnBoard.addElement(p);
	column++;
      }
    }

    moves = new Vector(80);
  }

  /*
   * setPieceAt receives a chess coordinate (e,g, 'a2' or 'b3') and
   * a piece and sets the piece at that board position.
   */
  void setPieceAt(char c, int r, Piece p){
    p.setBoard(this);
    p.setPosition(c,r);
    boardMatrix[8-r][(int) c - (int) 'a'] = p;
  }

  /*
   * getPieceAt receives a chess coordinate (e,g, 'a2' or 'b3') and
   * returns the  piece at that board position (which could be null).
   */
  Piece getPieceAt(char c, int r) throws ArrayIndexOutOfBoundsException{
    return boardMatrix[8-r][(int) c - (int) 'a'];
  }

  /* pieceAt receives a chess coordinate (e,g, 'a2' or 'b3') and
   * returns whether or not there is a piece in that square.
   */
  boolean pieceAt(char c, int r){
    return (getPieceAt(c,r) != null);
  }

  /*
   * getKing receives a color and returns the king of that color
   */
  Piece getKing(boolean color){
    return (color? wKing : bKing);
  }

  /**
   * movePieceOut takes a piece out of the board
   * pre:  piece must be on board
   */
  void movePieceOut(Piece q){
    Piece p = getPieceAt(q.getColumn(),q.getRow());
    //    System.err.println("Moving Piece Out:"+p);
    if (p.color()){  // White
      whitePiecesOut.addElement(p);
      whitePiecesOnBoard.removeElement(p);

    } else {  // Black
      blackPiecesOut.addElement(p);
      blackPiecesOnBoard.removeElement(p);
    }
  }

  /**
   * movePieceIn takes a piece in the board
   * pre:  piece must be out of board
   */
  void movePieceIn(Piece p){
    if (p.color()){  // White
      whitePiecesOnBoard.addElement(p);
      whitePiecesOut.removeElement(p);

    } else {  // Black
      blackPiecesOnBoard.addElement(p);
      blackPiecesOut.removeElement(p);
    }
  }

  /*
   * getWhitePiecesOnBoard returns the vector of white pieces in the board.
   */
  Vector getWhitePiecesOnBoard(){
    return whitePiecesOnBoard;
  }

  /*
   * getBlackPiecesOnBoard returns the vector of black pieces in the board.
   */
  Vector getBlackPiecesOnBoard(){
    return blackPiecesOnBoard;
  }


  /*
   * getWhitePiecesOut returns the vector of white pieces out of the board.
   */
  Vector getWhitePiecesOut(){
    return whitePiecesOut;
  }

  /*
   * getBlackPiecesOut returns the vector of black pieces out of the board.
   */
  Vector getBlackPiecesOut(){
    return blackPiecesOut;
  }

  /**
   * setCastling allows to enable/disable the generation of
   * castling as a move.
   */
  void setCastling(boolean castling){
    this.castling = castling;
  }

  /**
   * getCastling is used by Kings to decide whether to 
   * try to generate the castling as a move or not.
   */
  boolean getCastling(){
    return castling;
  }

  /*
   * generateMoves receives the color that plays next and 
   * returns a vector with the next possible valid moves.
   * DYNAMIC (changes board)
   */
  public Vector generateMoves(boolean color){
    //  This is TOO slow: (requires lookahead)
    return eliminateChecks(generateAllMoves(color), color);
    // New strategy: generate ALL moves, then in evaluation of
    // a static board, give a no-king position an INFINITE value,
    // and if there's no king in the board generateMoves -> null.
  }

  /*
   * generateAllMoves receives the color that plays next and 
   * returns a vector with the next possible valid moves,
   * including those that would create a check (invalid moves).
   * STATIC (does NOT change board)
   */
  public Vector generateAllMoves(boolean color){
    Vector possibleMoves = new Vector(50);
    // If there is no king --> no valid moves.
    if (!getKing(color).isOut()){   
      for (Enumeration pieces = 
	     (color? whitePiecesOnBoard.elements()
	      : blackPiecesOnBoard.elements()) ;
	   pieces.hasMoreElements();){ 
	// Create a new thread per Piece & Test performance...
	Piece p = (Piece) pieces.nextElement();
	Vector pieceMoves = p.generateMoves();
	// JDK 1.2: 
	// possibleMoves.addAll(pieceMoves);
	// JDK 1.1:
	for (Enumeration moves = pieceMoves.elements();
	     moves.hasMoreElements();){
	  possibleMoves.addElement(moves.nextElement());
	}
      }
    }
    return possibleMoves;
  }

  /*
   * generateMoveTree creates a tree of possible moves from a position.
   * receives the color that plays next,
   * receives the time it has allocated to "think" (-1 means unconstrained)
   * receives the maximum level to look ahead (<0 means unconstrained), 
   * returns a tree with the next possible valid moves.
   * DYNAMIC (changes board)
   */
  public MoveTree generateMoveTree(boolean color, long time, int level){
    Clock thinkingTime = new Clock();
    thinkingTime.start();

    int fn;
    try {
      fn = Algorithm.evaluateStaticFunction(this);
    } catch (CutSearchTreeException cste){
      return new MoveTree((Move) null, cste.getValue(), cste.getValue(), 
			  null, null);
    }
    if ((level == 0) || (time == 0)) 
      return new MoveTree((Move)null, fn, fn, null, null);

    if (level > 1) 
      System.err.print(".");

    // Old strategy: SLOW: check for validity of moves (no check)
    // Vector nextLevel = generateMoves(color);       // Generate next level
    // New strategy: generate ALL moves, then in evaluation of
    // a static board, give a no-king position an INFINITE value,
    // and if there's no king in the board generateMoves -> null.
    Vector nextLevel = generateAllMoves(color);       // Generate next level
    int bestValue = (color == Game.WHITE) ? -Game.INFINITY : Game.INFINITY;
    MoveTree bestMoveTree = null;
    MoveTree thisMoveTree = new MoveTree((Move)null, fn, 
					 bestValue, bestMoveTree, null);

    for (Enumeration moves = nextLevel.elements();
	 moves.hasMoreElements();){

      Move m = (Move) moves.nextElement();
                                           // for each move, generate tree
      doMove(m);
      long newTime = time-thinkingTime.value();
      MoveTree newMoves = generateMoveTree(!color, 
		       	    ((time == (long) -1.0)? (long) -1.0 : 
			     ((newTime > 0)?
			      newTime/nextLevel.size() : 0)),
		       	    level - 1);
      newMoves.setMove(m);
      thisMoveTree.addNextMove(newMoves);
      undoMove(m);

      /* Update best-level value and move at parent's tree */
      fn = newMoves.getMaxMin();
      if (color == Game.WHITE){             // choose maximum
	if (fn >= bestValue) {
	  bestValue = fn;
	  bestMoveTree = newMoves;
	}
      } else {                              // choose minimum
	if (fn <= bestValue) {
	  bestValue = fn;
	  bestMoveTree = newMoves;
	}
      }
    }
    
    thisMoveTree.setMaxMin(bestValue);
    thisMoveTree.setMaxMinTree(bestMoveTree);
    if (level > 2) 
      Game.debug(" bv=>"+bestValue+" bM==>"+bestMoveTree.getMove());
    return thisMoveTree;
  }

  /*
   * generateMoveTree creates a tree of possible moves from a position.
   * receives the color that plays next,
   * receives the time it has allocated to "think" (-1 means unconstrained)
   * receives the maximum level to look ahead (<0 means unconstrained), 
   * receives the alpha value to cut off minimizing plies,
   * receives the beta value to cut off maximizing plies,
   * returns a tree with the next possible valid moves.
   * DYNAMIC (changes board)
   */
  public MoveTree generateMoveTree(boolean color, long time, int level,
			    int alpha, int beta){
    Clock thinkingTime = new Clock();
    thinkingTime.start();

    int fn;
    try {
      fn = Algorithm.evaluateStaticFunction(this);
    } catch (CutSearchTreeException cste){
      return new MoveTree((Move) null, cste.getValue(), cste.getValue(), 
			  null, null);
    }
    if ((level == 0) || (time == 0)) 
      return new MoveTree((Move)null, fn, fn, null, null);

    if (level > 1) 
      System.err.print(".");

    Vector nextLevel = generateAllMoves(color);       // Generate next level
    int bestValue = (color == Game.WHITE) ? -Game.INFINITY : Game.INFINITY;
    MoveTree bestMoveTree = null;
    MoveTree thisMoveTree = new MoveTree((Move)null, fn, 
					 bestValue, bestMoveTree, null);

    for (Enumeration moves = nextLevel.elements();
	 moves.hasMoreElements();){

      Move m = (Move) moves.nextElement();
                                           // for each move, generate tree
      doMove(m);
      long newTime = time-thinkingTime.value();
      MoveTree newMoves = generateMoveTree(!color, 
		       	    ((time == (long) -1.0)? (long) -1.0 : 
			     ((newTime > 0)?
			      newTime/nextLevel.size() : 0)),
		       	    level - 1, alpha, beta);
      newMoves.setMove(m);
      // temporarily, we won't keep track of tree structure...
      // just the best move.  Later, keep the best moveS here...
      // thisMoveTree.addNextMove(newMoves);
      undoMove(m);

      /* Update best-level value and move at parent's tree */
      fn = newMoves.getMaxMin();
      if (color == Game.WHITE){             // choose maximum
	if (fn >= bestValue) {
	  bestValue = fn;
	  bestMoveTree = newMoves;
	  if (bestValue > alpha)            // update threshold
	    alpha = bestValue;
	}
	if (bestValue > beta)
	  break;                            // cut tree search
      } else {                              // choose minimum
	if (fn <= bestValue) {
	  bestValue = fn;
	  bestMoveTree = newMoves;
	  if (bestValue < beta)             // update threshold
	    beta = bestValue;
	}
	if (bestValue < alpha)
	  break;                            // cut tree search
      }
    }
    
    thisMoveTree.setMaxMin(bestValue);
    thisMoveTree.setMaxMinTree(bestMoveTree);
    if (level > 2) 
      Game.debug(" bv=>"+bestValue+" bM==>"+bestMoveTree.getMove());
    return thisMoveTree;
  }


  /*
   * generateMoveTree creates a tree of possible moves from a position.
   * receives the color that plays next,
   * receives the time it has allocated to "think" (-1 means unconstrained)
   * receives the maximum level to look ahead (<0 means unconstrained), 
   * receives the depth of the tree to generate (doesn't change),
   * receives the alpha value to cut off minimizing plies,
   * receives the beta value to cut off maximizing plies,
   * receives the vector of next moves to evaluate,
   * returns a tree with the next possible valid moves.
   * DYNAMIC (changes board)
   */
  public MoveTree generateMoveTree(boolean color, long time, int level, int depth,
			    int alpha, int beta, Vector nextLevel){
    Clock thinkingTime = new Clock();
    thinkingTime.start();

    int fn;
    try {
      fn = Algorithm.evaluateStaticFunction(this);
    } catch (CutSearchTreeException cste){
      return new MoveTree((Move) null, cste.getValue(), cste.getValue(), 
			  null, null);
    }
    if ((level == 0) || (time == 0)) 
      return new MoveTree((Move)null, fn, fn, null, null);

    if (Game.debug)
      if (level > 1) 
	System.err.print(".");

    if (nextLevel == null)
      nextLevel = generateAllMoves(color);       // Generate next level
    int bestValue = (color == Game.WHITE) ? -Game.INFINITY : Game.INFINITY;
//     MoveTree bestMoveTree = null;
    MoveTree thisMoveTree = new MoveTree((Move)null, fn, 
					 bestValue, null, null);

    for (Enumeration moves = nextLevel.elements();
	 moves.hasMoreElements();){

      Move m = (Move) moves.nextElement();
                                           // for each move, generate tree
      doMove(m);
      long newTime = time-thinkingTime.value();
      MoveTree newMoves = 
	iterativeDeepening(!color, 
			   ((time == (long) -1.0)? (long) -1.0 : 
			    ((newTime > 0)?
			     newTime/nextLevel.size() : 0)),
			   level - 1,
			   alpha, beta, 
			   0);
      newMoves.setMove(m);
      fn = newMoves.getMaxMin();
      if ((color? (fn>=Game.INFINITY):(fn<=-Game.INFINITY)) &&
	  (newMoves.depth() == 3) &&
	  (!isCheck(!color))){
	// Apparent mate... but it may actually be a drawn king!!
	fn = 0;
	newMoves.setMaxMin(0);
      }
      if (level == depth)                  // keep next level of moves sorted.
	thisMoveTree.addNextMoveSorted(color, newMoves);
      undoMove(m);

      /* Update best-level value and move at parent's tree */
      if (color == Game.WHITE){             // choose maximum
	if (fn > bestValue) {
	  bestValue = fn;
// 	  bestMoveTree = newMoves;
	  if (bestValue > alpha)            // update threshold
	    alpha = bestValue;
	}
	if (bestValue >= beta){
	  break;                            // cut tree search
	}
      } else {                              // choose minimum
	if (fn < bestValue) {
	  bestValue = fn;
// 	  bestMoveTree = newMoves;
	  if (bestValue < beta)             // update threshold
	    beta = bestValue;
	}
	if (bestValue <= alpha){
	  break;                            // cut tree search
	}
      }
    }
    
    thisMoveTree.setMaxMin(bestValue);
    try {
      thisMoveTree.setMaxMinTree( (MoveTree) 
				  thisMoveTree.getNextMoves().elementAt(0));
    } catch (NullPointerException noTree){};
    if (level > 2) 
      Game.debug(" bv=>"+bestValue+" bM==>"+
		 ((thisMoveTree.getMaxMinTree() != null)? 
		  thisMoveTree.getMaxMinTree().getMove().toString() : "null"));
    return thisMoveTree;
  }


  /** eliminateChecks takes a vector of all possible moves and
   * prunes out those that would create a check for the 
   * given color.
   * DYNAMIC (changes board)
   */
   Vector eliminateChecks(Vector moves, boolean color){
     Vector newMoves = new Vector(moves.size());
     for(Enumeration e = moves.elements();
	 e.hasMoreElements();){
       Move m = (Move) e.nextElement();
       doMove(m);
       if (!isCheck(color)) // move is ok.
	 newMoves.addElement(m);
       undoMove(m);
     }
     return newMoves;
   }

  /*
   * isCkechMateOrDraw receives a color and returns whether or not
   * that color is in checkmate.
   */
  boolean isCheckMateOrDraw(boolean color){
    return (isCheckMate(color) || (isDraw(color)));
  }

  /*
   * isCkechMate receives a color and returns whether or not
   * that color is in checkmate.
   */
  boolean isCheckMate(boolean color){
    Piece k = getKing(color);
    return (isCheck(color) && (generateMoves(color).size() == 0));
  }

  /*
   * isDraw receives a color and returns whether or not
   * that color is in draw.
   */
  boolean isDraw(boolean color){
    Piece k = getKing(color);
    return ((!isCheck(color)) && (generateMoves(color).size() == 0));
    // Incomplete:  Must check no repetition of moves; king-knight vs king,...
  }

  /**
   * isCheck receives a color and returns whether or not
   * that color is in check.
   */
  boolean isCheck(boolean color){
    // check if the king of this color is attacked.
    Piece k = getKing(color);
    return (k.isAttacked());
  }

  /* 
   * positionAttacked receives a color and a board coordinate (col, row)
   * and returns whether or not that color is attacking the given position.
   */
  boolean positionAttacked(boolean color, char c, int r){
    try {
      Piece attackedPiece = getPieceAt(c,r);
      if (attackedPiece.color() != color) {
	for (Enumeration pieces =
	       (color? whitePiecesOnBoard.elements()
		: blackPiecesOnBoard.elements()) ;
	     pieces.hasMoreElements();){
	  Piece p = (Piece) pieces.nextElement();
	  if (p.attacks(attackedPiece)) 
	    return true;
	}
      }
    } catch (IndexOutOfBoundsException e){}
    return false;
  }

  /* 
   * positionReachable receives a color and a board coordinate (col, row)
   * and returns whether or not that color reaches the given position.
   * reach == potential/real attack.
   */
  boolean positionReachable(boolean color, char c, int r){
    for (Enumeration pieces =
	   (color? whitePiecesOnBoard.elements()
	         : blackPiecesOnBoard.elements()) ;
	 pieces.hasMoreElements();){
      Piece p = (Piece) pieces.nextElement();
      if (p.reaches(c,r)) 
	return true;
    }
    return false;
  }

  /*
   * doMove receives a Move object and performs the move on this
   * board.
   */
  public void doMove(submit.Move m2){
    if (!(m2 instanceof Move)) return;
    Move m=(Move) m2;
    // If king or rook, record No of moves for castling
    Piece movingPiece = m.getMovingPiece();
    if (movingPiece instanceof King)
      ((King)movingPiece).incMoves();
    else if (movingPiece instanceof Rook)
      ((Rook)movingPiece).incMoves();

    if (m.isCastling()){
      int row = m.getFinalRow();
      // Move king...
      setPieceAt(m.getFinalColumn(), row, movingPiece);
      boardMatrix[8-row][(int) 'e' - (int) 'a'] = null;
      // Move rook...
      if (m.getFinalColumn()=='g'){ // Short castling
	setPieceAt('f', row, getPieceAt('h', row));
	boardMatrix[8-row][(int) 'h' - (int) 'a'] = null;
      } else {                      // Long castling
	setPieceAt('d', row, getPieceAt('a', row));
	boardMatrix[8-row][(int) 'a' - (int) 'a'] = null;
      }
    } else {
      /* Remove displaced piece */
      Piece displacedPiece = m.getDisplacedPiece();
      if (displacedPiece != null){
	movePieceOut(displacedPiece);
	if (m.isPawnInWay()){
	  boardMatrix[8-displacedPiece.getRow()]
	    [(int) displacedPiece.getColumn() - (int) 'a'] = null;
	} 
      }
      /* Update piece and board from initial to final position */
      Piece newPiece = m.getNewPiece();
      if (newPiece != null){  // pawn-to-last-row
	movePieceOut(movingPiece);
	setPieceAt(m.getFinalColumn(), m.getFinalRow(), newPiece);
	movePieceIn(newPiece);
      }
      else
	setPieceAt(m.getFinalColumn(), m.getFinalRow(), movingPiece);
      /* Update board's initial position */
      boardMatrix[8-m.getInitialRow()]
	[(int) m.getInitialColumn() - (int) 'a'] = null;
    }
    moves.addElement(m);
  }

  /*
   * undoMove receives a Move object representing the last move 
   * done to the current Board and undoes the move.
   */
  public void undoMove(submit.Move m2){
    if (!(m2 instanceof Move)) return;
    Move m=(Move) m2;
    // If king or rook, record No of moves for castling
    Piece movingPiece = m.getMovingPiece();
    if (movingPiece instanceof King)
      ((King)movingPiece).decMoves();
    else if (movingPiece instanceof Rook)
      ((Rook)movingPiece).decMoves();

    if (m.isCastling()){
      int row = m.getFinalRow();  // == m.getInitialRow()
      // Move king back...
      setPieceAt(m.getInitialColumn(), row, movingPiece);
      boardMatrix[8-row][(int) m.getFinalColumn() - (int) 'a'] = null;
      // Move rook back...
      if (m.getFinalColumn()=='g'){ // Short castling
	setPieceAt('h', row, getPieceAt('f', row));
	boardMatrix[8-row][(int) 'f' - (int) 'a'] = null;
      } else {                      // Long castling
	setPieceAt('a', row, getPieceAt('d', row));
	boardMatrix[8-row][(int) 'd' - (int) 'a'] = null;
      }
    } else {
       /* Update piece and board from final to initial position */
       setPieceAt(m.getInitialColumn(), m.getInitialRow(), movingPiece);
       Piece newPiece = m.getNewPiece();
       if (newPiece != null){  // pawn-to-last-row: restablish piece lists.
	 movePieceIn(movingPiece);
	 if (newPiece.color())  // White
	   whitePiecesOnBoard.removeElement(newPiece);
	 else  // Black
	   blackPiecesOnBoard.removeElement(newPiece);
       }
       /* Restore displaced piece or clear final position */
       Piece displacedPiece = m.getDisplacedPiece();
       if (displacedPiece != null){
	 movePieceIn(displacedPiece);
	 boardMatrix[8-displacedPiece.getRow()]
	   [(int) displacedPiece.getColumn() - (int) 'a'] = displacedPiece;
	 if (m.isPawnInWay())
	   boardMatrix[8-m.getFinalRow()]
	     [(int) m.getFinalColumn() - (int) 'a'] = null;
       } else {
       /* Update board's final position */
	 boardMatrix[8-m.getFinalRow()]
	   [(int) m.getFinalColumn() - (int) 'a'] = null;
       }
     } 
    moves.removeElementAt(moves.size()-1);  // remove last move.
  }

  /**
   * lastMove returns the last move made in this board
   */
  public Move lastMove(){
    return ((Move) moves.lastElement());
  }

  /**
   * moveNumber returns the move number n, starting from 0.
   */
  public Move moveNumber(int n){
    return ((Move) moves.elementAt(n));
  }

  /**
   * moveRepetition returns whether the last moves in the board
   * have been repetitions or not, for both players.
   * For example, if the moves are:
   *
   *     7.   b3-a4 b7-c8
   *     8.   a4-b3 c8-b7
   *     9.   b3-a4 b7-c8
   *
   * Then this function returns true for move 9.
   */
  // There may be larger cycles to be checked (?)
  public boolean moveRepetition(){
    // Check the last five moves:  m4, m3, m2, m1, m
    // return ((m == m4) && (inverse(m,m2)) && (inverse(m1,m3)))
    try {
      int lastMoveIndex = moves.size()-1;
      Move m = lastMove();
      Move m4 = moveNumber(lastMoveIndex-4);
      if (m.equals(m4)){
	Move m2 = moveNumber(lastMoveIndex-2);
	if (m.inverseMove(m2)){
	  Move m3 = moveNumber(lastMoveIndex-3);
	  Move m1 = moveNumber(lastMoveIndex-1);
	  return m1.inverseMove(m3);
	}
      }
    } catch (IndexOutOfBoundsException e){ // No such move
    } catch (NoSuchElementException e){}    // No moves
    return false;
  }

  public Game getGame(){
    return game;
  }

  public void printAscii(){
    System.out.println("***** C H E S S *****");
    for (int j=0; j < 8; j++) System.out.print("----");
    System.out.println();
    for (int i=0; i < 8; i++){
      System.out.print("| ");
      for (int j=0; j<8; j++){
	Piece p;
	if (game.computerColor()) 
	  p = boardMatrix[7-i][7-j];
	else
	  p = boardMatrix[i][j];
	if (p != null)
	  System.out.print(p);
	else System.out.print(" ");
	System.out.print(" | ");
      }
      System.out.println();
      for (int j=0; j<8; j++) System.out.print("----");
      System.out.println();
    }
  }

  public MoveTree iterativeDeepening(boolean color, int depth) {
    return iterativeDeepening(color,-1,depth,-submit.constants.INFINITY, submit.constants.INFINITY, 0);
  }
  /**
   * iterativeDeepening creates a tree of possible moves from a position.
   * receives the color that plays next,
   * receives the time it has allocated to "think" (-1 means unconstrained)
   * receives the maximum depth to look ahead (<0 means unconstrained),
   * receives the alpha value to cut off minimizing plies,
   * receives the beta value to cut off maximizing plies,
   * receives the X window value to prune next moves to evaluate,
   * returns a tree with the next possible valid moves.
   * DYNAMIC (changes board)
   */
  MoveTree iterativeDeepening(boolean color, long time, int depth,
			      int alpha, int beta, int xwindow){
    Clock thinkingTime = new Clock();
    thinkingTime.start();

    int fn;
    try {
      fn = Algorithm.evaluateStaticFunction(this);
    } catch (CutSearchTreeException cste){
      return new MoveTree((Move) null, cste.getValue(), cste.getValue(),
			  null, null);
    }
    if ((depth == 0) || (time == 0))
      return new MoveTree((Move)null, fn, fn, null, null);

    MoveTree currentMoveTree = null;
    Vector nextMoves = generateAllMoves(color);
    int level = 1;
    while ((depth > 0) && (level <= depth)){
      if (depth > 4)
	Game.debug("L"+level+nextMoves);
      long newTime = time-thinkingTime.value();
      if ((time != -1.0) && (newTime <= 0)){
	game.setLevel(level);
	break;
      }
      currentMoveTree =
	generateMoveTree(color,
			 ((time == (long) -1.0)? (long) -1.0 :
			  ((newTime > 0)? newTime : 0)),
			 level,
			 level,  //depth?
			 ((level == depth)? alpha: -Game.INFINITY),
			 ((level == depth)? beta: Game.INFINITY),
			 nextMoves);
      nextMoves = currentMoveTree.prune(xwindow);
      level++;
    }
    return currentMoveTree;
  }

}
