/*
 * Chess game implementation.
 * By Carlos Varela.  February, 1997
 *
 * King class.  Contains a chess piece representing a king.
 */

package chess;

import java.util.*;

class King extends Piece implements java.io.Serializable {

  int noMoves = 0;  //  Can castle only if it hasn't moved.

  King(boolean col){
    value = (col? Game.INFINITY : -Game.INFINITY);
    color = col;
  }

  boolean validMove(Board b, boolean col, char c, int r){
    try {
      Piece p;
      return (((p = b.getPieceAt(c,r)) == null) || // A piece NOT in the way
	      (p.color() != col));                 // Or opposite color
// For Efficiency:  Allow check generating moves,
//                  and let search get rid of this bad move.
//	    && (!b.positionReachable(!col,c,r)));  // And not creates check...
    } catch (ArrayIndexOutOfBoundsException e){    // Out of board.
      return false;
    }
  }

  Vector generateMoves(){
    Vector kingMoves = new Vector(8);
    char c = getColumn();
    int r = getRow();
    Board b = getBoard();
    boolean col = color();
    int ri;
    char ci;
    // King movements
    
                                 // Right Upward move 
    ri = r + 1;
    ci = (char) (c + 1);
    if (validMove(b,col,ci,ri))
      kingMoves.addElement(new Move(this, c, r, ci, ri));

                                 // Right Downward move 
    ri = r - 1;
    ci = (char) (c + 1);
    if (validMove(b,col,ci,ri))
      kingMoves.addElement(new Move(this, c, r, ci, ri));

                                 // Left Upward move 
    ri = r + 1;
    ci = (char) (c - 1);
    if (validMove(b,col,ci,ri))
      kingMoves.addElement(new Move(this, c, r, ci, ri));

                                 // Left Downward move 
    ri = r - 1;
    ci = (char) (c - 1);
    if (validMove(b,col,ci,ri))
      kingMoves.addElement(new Move(this, c, r, ci, ri));

                                 // Upward move 
    ri = r + 1;
    ci = c;
    if (validMove(b,col,ci,ri))
      kingMoves.addElement(new Move(this, c, r, ci, ri));

                                 // Downward move 
    ri = r - 1;
    ci = c;
    if (validMove(b,col,ci,ri))
      kingMoves.addElement(new Move(this, c, r, ci, ri));

                                 // Leftward move 
    ri = r;
    ci = (char) (c - 1);
    if (validMove(b,col,ci,ri))
      kingMoves.addElement(new Move(this, c, r, ci, ri));

                                 // Rightward move 
    ri = r;
    ci = (char) (c + 1);
    if (validMove(b,col,ci,ri))
      kingMoves.addElement(new Move(this, c, r, ci, ri));

   if (b.getCastling()){
                                 // 0-0: Short castling
      try {
	//  We need to check:
	// (1) King has not moved and is not in check.
	// (2) Rook in column "h" has not moved.
	// (3) Positions f[18] and g[18] are empty. (1:W, 8:B)
	// (4) Positions f[18] and g[18] are not reachable. (1:W, 8:B)
	// If possible, we'll represent the move as:
	//      ("K",'e',1,'g',1) for W
	// and  ("k",'e',8,'g',8) for B.
	// It should be printed as "0-0".
	if (
	    (!hasMoved()) &&                                // (1)
	    (!b.pieceAt('f',r)) && (!b.pieceAt('g',r)) &&   // (3)
	    (!((Rook) b.getPieceAt('h', r)).hasMoved()) && // (2)
	    //  FIX Efficiency: these next three in ONE method...
	    // overload generateAllMoves(boolean color, Method stop);
	    // and give it a stop method that checks these three.
	    (!isAttacked()) &&                              // (1)
	    (!b.positionReachable(!col,'f',r)) &&           // (4)
	    (!b.positionReachable(!col,'g',r))              // (4)
	    )
	  kingMoves.addElement(new Move(this, c, r, 'g', r));
      } catch (NullPointerException e){  // Not a piece in 'h' column
      } catch (ClassCastException e){    // Not a rook in 'h' column
      } catch (ArrayIndexOutOfBoundsException e){};  // Out of board

      // 0-0-0: Long castling 
      try {
	//  We need to check:
	// (1) King has not moved and is not in check.
	// (2) Rook in column "a" has not moved.
	// (3) Positions b[18], c[18] and d[18] are empty. (1:W, 8:B)
	// (4) Positions c[18] and d[18] are not attacked. (1:W, 8:B)
	// If possible, we'll represent the move as:
	//      ("K",'e',1,'c',1) for W
	// and  ("k",'e',8,'c',8) for B.
	// It should be printed as "0-0-0".
	if (
	    (!hasMoved()) &&                                // (1)
	    (!b.pieceAt('b',r)) && (!b.pieceAt('c',r)) && 
	    (!b.pieceAt('d',r)) &&                          // (3)
	    (!((Rook) b.getPieceAt('a', r)).hasMoved()) && // (2)
	    (!isAttacked()) &&                              // (1)
	    (!b.positionReachable(!col,'c',r)) &&           // (4)
	    (!b.positionReachable(!col,'d',r))              // (4)
	    )
	  kingMoves.addElement(new Move(this, c, r, 'c', r));
      } catch (NullPointerException e){  // Not a piece in 'a' column
      } catch (ClassCastException e){    // Not a rook in 'a' column
      } catch (ArrayIndexOutOfBoundsException e){};  // Out of board
    }

    return kingMoves;
  }

  boolean isProtected(){
    // Incomplete.
    return true;
  }

  void incMoves(){
    noMoves++;
  }

  void decMoves(){
    noMoves--;
  }

  boolean hasMoved(){
    return (noMoves > 0);
  }

  public String toString(){
    return ((color == Game.WHITE)? "K" : "k");
  }

}


