/*
 * Chess game implementation.
 * By Carlos Varela.  February, 1997
 *
 * Move class.  Contains a valid move in the game.
 */

package chess;

import java.io.*;

public class Move implements submit.Move, java.io.Serializable {

  Piece movingPiece;
  char initialColumn;
  int initialRow;
  int finalRow;
  char finalColumn;
  Piece displacedPiece;
  boolean pawnInWay;
  Piece newPiece;       // when a pawn gets to the last row

  /**
   * Constructs a move and infers from the board if there is
   * a piece being displaced.
   */
  Move(Piece p, char ic, int ir, char fc, int fr) {
    movingPiece = p;
    initialColumn = ic;
    initialRow = ir;
    finalColumn = fc;
    finalRow = fr;
    displacedPiece = p.getBoard().getPieceAt(fc, fr);
    pawnInWay = false;
    newPiece = null;
  }

  /**
   * Constructs a move with the displaced pawn in the way.
   */
  Move(Piece p, char ic, int ir, char fc, int fr, Piece dp) {
    movingPiece = p;
    initialColumn = ic;
    initialRow = ir;
    finalColumn = fc;
    finalRow = fr;
    displacedPiece = dp;  
    pawnInWay = true;
    newPiece = null;
  }

  /**
   * Constructs a move with the piece replacing the last-row pawn.
   */
  Move(Piece p, char ic, int ir, char fc, int fr, Piece dp, Piece np) {
    movingPiece = p;
    initialColumn = ic;
    initialRow = ir;
    finalColumn = fc;
    finalRow = fr;
    displacedPiece = dp;
    pawnInWay = false;
    newPiece = np;
  }

  /**
   * Constructs a move from its String representation.
   */
  Move(Board b, String move){
    getMove(b, move);
  }

  /**
   * Constructs a move reading it from standard input.
   */
  Move(Board b){
    try {
      System.out.print("Your move (e.g. e2e4): ");
      MoveReader min = new MoveReader(
				      new BufferedReader(new InputStreamReader(System.in)));
      String move = min.readMove();
      getMove(b, move);
    } catch (Exception e) {
      movingPiece = null;  /* Invalid move */
    }
  }

  /**
   * Receives a string "e2e4 ..." and returns the
   * move that corresponds to the first 4 characters.
   * If a pawn gets to the last row, read the 5th char, e.g: e7e8q
   * Throws an IndexOutOfBoundsException if wrong String format.
   */
  void getMove(Board b, String move){
    if ((initialColumn = move.charAt(0)) == '0'){
      // "0-0  " or "0-0-0"
      initialColumn = 'e';
      if (b.getGame().nextPlayer()){  // Plays white
	initialRow = finalRow = 1;
	movingPiece = b.getKing(Game.WHITE);
      } else {                        // Plays black
	initialRow = finalRow = 8;
	movingPiece = b.getKing(Game.BLACK);
      }
      if ((move.length() > 5) && (move.substring(0,5).equals("0-0-0"))){
	finalColumn = 'c';
      } else if (move.substring(0,3).equals("0-0")){
	finalColumn = 'g';
      } else throw new StringIndexOutOfBoundsException();
    } else {
      initialRow = move.charAt(1)-'0';
      finalColumn = move.charAt(2);
      finalRow = move.charAt(3)-'0';
      movingPiece = b.getPieceAt(initialColumn, initialRow);
      displacedPiece = b.getPieceAt(finalColumn, finalRow);
      pawnInWay = false;
      if ((movingPiece instanceof Pawn) &&
	  (initialColumn != finalColumn) &&
	  (displacedPiece == null) &&
	  (((movingPiece.color() == Game.WHITE) && (initialRow == 5)) || 
	   ((movingPiece.color() == Game.BLACK) && (initialRow == 4)))){
	// pawn-in-the-way...
	displacedPiece = b.getPieceAt(finalColumn,
				      (movingPiece.color()? 5 : 4));
	pawnInWay = true;
      }
      newPiece = null;
      if ((movingPiece instanceof Pawn) && 
	  ((finalRow == 8) || (finalRow == 1))){  
	// pawn in last row
	switch (move.charAt(4)){
	case 'Q':
	case 'q':
	  newPiece = new Queen(b.getGame().nextPlayer());
	  break;
	case 'R':
	case 'r':
	  newPiece = new Rook(b.getGame().nextPlayer());
	  break;
	case 'B':
	case 'b':
	  newPiece = new Bishop(b.getGame().nextPlayer());
	  break;
	case 'N':
	case 'n':
	  newPiece = new Knight(b.getGame().nextPlayer());
	  break;
	default:
	  System.err.println("Invalid piece:"+move.charAt(4));
	  throw new StringIndexOutOfBoundsException();
	}
      }
    }
    return;
  }

  public boolean getColor(){
    return movingPiece.color();
  }

  public Piece getMovingPiece(){
    return movingPiece;
  }

  public char getInitialColumn(){
    return initialColumn;
  }

  public int getInitialRow(){
    return initialRow;
  }

  public char getFinalColumn(){
    return finalColumn;
  }

  public int getFinalRow(){
    return finalRow;
  }

  public Piece getDisplacedPiece(){
    return displacedPiece;
  }

  public Piece getNewPiece(){
    return newPiece;
  }

  public boolean validMove(Board b, boolean color)
    throws CheckMateException, DrawException {
    boolean valid = ((movingPiece != null) &&
		     (movingPiece.color() == color) &&
		     (b.generateMoves(color).contains(this)));
    if (!valid)
      if (b.generateMoves(color).size() == 0)
	if (b.isCheck(color))
	  throw new CheckMateException(!color);
	else throw new DrawException();
    return valid;
  }

  public boolean isCastling(){
    return ((movingPiece instanceof King) &&
	    (initialColumn == 'e') &&
	    ((finalColumn == 'g') || (finalColumn == 'c')));
  }

  public boolean isPawnInWay(){
    return pawnInWay;
  }

  /**
   * inverseMove returns true if <code>this</code> is the inverse of
   * the given <b>immediate predecessor</b> move of the same color.
   */
  public boolean inverseMove(Move move){
    // We assume it is the same movingPiece and it is not a Pawn
    // (because pawns can't move backwards) therefore no newPiece/pawnInWay.
    return ((initialColumn == move.getFinalColumn()) &&
	    (initialRow == move.getFinalRow()) &&
	    (finalColumn == move.getInitialColumn()) &&
	    (finalRow == move.getInitialRow()) &&
	    (move.getDisplacedPiece() == null) &&
	    (displacedPiece == null));
  }

  public boolean equals(Object o){
    boolean eq = 
      ((o.getClass() == this.getClass())
       && (((Move)o).movingPiece == this.movingPiece)
       && (((Move)o).initialColumn == this.initialColumn)
       && (((Move)o).initialRow == this.initialRow)
       && (((Move)o).finalColumn == this.finalColumn)
       && (((Move)o).finalRow == this.finalRow)
       // May not be exactly the same newPiece object, but
       // it represents the same Piece and therefore Move.
       && (((((Move)o).newPiece == null) && (this.newPiece == null)) ||
	   ((Move)o).newPiece.getValue() == (this.newPiece.getValue())) 
       && (((Move)o).displacedPiece == this.displacedPiece));
    return eq;
  }

  public String toString(){
    // Check for castling
    if (isCastling()){
      return("0-0"+((finalColumn == 'c')?"-0":"")+'['+movingPiece+']');
    }
    return(initialColumn+Integer.toString(initialRow)+
	   // 	   ((displacedPiece == null)? '-' : 'x')+
	   finalColumn+Integer.toString(finalRow)+
	   ((newPiece != null)? newPiece.toString() : "")
	   //       +'['+((movingPiece != null)? movingPiece.toString() : " ")+']'
	   );
  }

}
