/*
 * Chess game implementation.
 * By Carlos Varela.  February, 1997
 *
 * Piece class.  Contains an abstraction of a chess piece.
 */

package chess;

import java.util.*;

abstract class Piece implements java.io.Serializable {

  Board board;
  boolean color;
  char column;
  int row;
  int value;

  boolean color(){
    return color;
  }

  void setBoard(Board b){
    board = b;
  }
  
  Board getBoard(){
    return board;
  }

  void setPosition(char c, int r){
    column = c;
    row = r;
  }

  char getColumn(){
    return column;
  }

  int getRow(){
    return row;
  }

  /** generateMoves returns a Vector of all
   * possible Moves for this piece in the current
   * board position.
   */
  abstract Vector generateMoves();

  /** attacks returns true if and only if
   * in the current board position, <code>this</code> piece
   * can displace the given piece of the opponent.
   */
  boolean attacks(Piece p){
    board.setCastling(false); // Castling will not "displace" any piece
                              // and produces an infinite loop... 
    Vector nextMoves = generateMoves();
    board.setCastling(true);
    for (Enumeration moves = nextMoves.elements();
	 moves.hasMoreElements();){
      Move m = (Move) moves.nextElement();
      if (p.equals(m.getDisplacedPiece()))
	return true;
    }
    return false;
  }

  /** attacks returns true if and only if
   * in the current board position, this piece
   * can move to the position (c,r),
   * displacing a piece of the opponent.
   */
  boolean attacks(char c, int r){
    try {
      // if there is a piece to attack
      if (board.getPieceAt(c,r).color() != color) 
	return reaches(c,r);
    } catch (IndexOutOfBoundsException e){}
    return false;
  }

  /** reaches returns true if and only if
   * in the current board position, this piece
   * can move to the position (c,r).
   * Notice that a piece doesn't "reach" its current position.
   */
  boolean reaches(char c, int r){
    board.setCastling(false); // Castling will not "reach" any position
                              // and produces an infinite loop... 
    Vector nextMoves = generateMoves();
    board.setCastling(true);
    for (Enumeration moves = nextMoves.elements();
	 moves.hasMoreElements();){
      Move m = (Move) moves.nextElement();
      if ((m.getFinalRow() == r) && (m.getFinalColumn() == c))
	return true;
    }
    return false;
  }

  /** defends returns true if and only if
   * in the current board position, this piece
   * could move to the position (c,r),
   * where there is currently a piece of the same color.
   */
  boolean defends(char c, int r){
    // This is where "generateMoves" *usually* stops...
    return false;
  }

  boolean isAttacked(){
    // generate all possible moves for other player and
    // see if one of them can kill this piece.
    // STATIC (does NOT change board)
    // FIX efficiency:  
    //   See if any piece of other color can kill this.
    //   It will not generate ALL moves, but instead, it will 
    //   stop as soon as an attacking piece is discovered.
    board.setCastling(false); // Castling will not kill any piece
                              // and produces an infinite loop... 
    Vector otherMoves = board.generateAllMoves(!color);
    board.setCastling(true);
    for (Enumeration moves = otherMoves.elements();
	 moves.hasMoreElements();){
      Move m = (Move) moves.nextElement();
      if (this.equals(m.getDisplacedPiece()))
	return true;
    }
    return false;
  }

  boolean isProtected(){
    return true;
  }

  int getValue(){
    return value;
  }

  boolean isCentered(){
    return (((column == 'd') || (column == 'e')) &&
	    ((row == 4) || (row == 5)));
  }

  /* there should be attacksCenter()/defendsCenter() too */
  boolean reachesCenter(){
    board.setCastling(false); // Castling will not "reach" any position
                              // and produces an infinite loop... 
    Vector nextMoves = generateMoves();
    board.setCastling(true);
    for (Enumeration moves = nextMoves.elements();
	 moves.hasMoreElements();){
      Move m = (Move) moves.nextElement();
      int r = m.getFinalRow();
      char c = m.getFinalColumn();
      if (((r == 4) || (r == 5)) &&
	  ((c == 'd') || (c == 'e')))
	return true;
    }
    return false;
  }

  boolean attacksCenter(){
    board.setCastling(false); // Castling will not "attack" any position
                              // and produces an infinite loop... 
    Vector nextMoves = generateMoves();
    board.setCastling(true);
    for (Enumeration moves = nextMoves.elements();
	 moves.hasMoreElements();){
      Move m = (Move) moves.nextElement();
      int r = m.getFinalRow();
      char c = m.getFinalColumn();
      if (((r == 4) || (r == 5)) &&
	  ((c == 'd') || (c == 'e')) &&
	  (m.getDisplacedPiece() != null))
	return true;
    }
    return false;
  }

  boolean makesCheck(){
    return attacks(board.getKing(!color));
  }

  boolean isOut(){
    return (color? board.getWhitePiecesOut().contains(this)
	         : board.getBlackPiecesOut().contains(this));
  }

  public boolean equals(Object o){
    boolean eq = 
      ((o != null)
       && (o.getClass() == this.getClass())
       && (((Piece)o).board == this.board)
       && (((Piece)o).color == this.color)
       && (((Piece)o).column == this.column)
       && (((Piece)o).row == this.row));
       // if same piece class and color -> same value.
    return eq;
  }

  public String toString(){
    return ("Piece");
  }

}




